"""Molecular rendering."""

from moldoc._internal.molecule import (
    Atom,
    AtomConfig,
    Bond,
    Color,
    Material,
    MeshLambertMaterial,
    MeshNormalMaterial,
    MeshPhongMaterial,
    MeshPhysicalMaterial,
    MeshStandardMaterial,
    Molecule,
    MoleculeConfig,
)

__all__ = [
    "Atom",
    "AtomConfig",
    "Bond",
    "Color",
    "Material",
    "MeshLambertMaterial",
    "MeshNormalMaterial",
    "MeshPhongMaterial",
    "MeshPhysicalMaterial",
    "MeshStandardMaterial",
    "Molecule",
    "MoleculeConfig",
]



from moldoc._internal.javascript.bool import bool_to_javascript
from moldoc._internal.javascript.color import color_to_javascript
from moldoc.molecule import MoleculeConfig


def get_scene_config(
    container_id: str,
    molecule_config: MoleculeConfig | None,
) -> str:
    scene_config = {
        "containerId": f'"{container_id}"',
        **_get_scene_config(molecule_config),
    }

    return ",".join(f"{key}:{value}" for key, value in scene_config.items())


def _get_scene_config(config: MoleculeConfig | None) -> dict[str, str]:
    result: dict[str, str] = {}

    if config is None:
        return result

    background_color = config.get_background_color()
    if background_color is not None:
        result["backgroundColor"] = color_to_javascript(
            color=background_color,
        )

    is_outlined = config.is_outlined()
    if is_outlined is not None:
        result["outline"] = bool_to_javascript(is_outlined)

    return result



import pkgutil
import typing

from docutils import nodes
from sphinx.application import Sphinx
from sphinx.environment import BuildEnvironment
from sphinx.util.docutils import SphinxDirective
from sphinx.writers.html5 import HTML5Translator

from moldoc._internal.javascript.atoms import get_atom_array
from moldoc._internal.javascript.bonds import get_bond_array
from moldoc._internal.javascript.mesh_config import get_mesh_config
from moldoc._internal.javascript.scene_config import get_scene_config
from moldoc.molecule import Molecule, MoleculeConfig
from moldoc.version import __version__


class MolDocNode(nodes.Body, nodes.Element):
    def __init__(
        self,
        moldoc_name: str,
        molecule: Molecule,
    ) -> None:
        super().__init__()
        self._moldoc_name = moldoc_name
        self._molecule = molecule

    def get_molecule(self) -> Molecule:
        return self._molecule

    def get_moldoc_name(self) -> str:
        return self._moldoc_name


class MolDoc(SphinxDirective):
    has_content = True

    def run(self) -> list[MolDocNode]:
        content = "\n".join(self.content)
        globals_: dict[str, typing.Any] = {}
        exec(content, globals_)  # noqa: S102
        node = MolDocNode(
            moldoc_name=f'moldoc_{self.env.new_serialno("moldoc")}',
            molecule=globals_["moldoc_display_molecule"],
        )

        if not hasattr(self.env, "moldoc_documents"):
            self.env.moldoc_documents = set()  # type: ignore[attr-defined]

        self.env.moldoc_documents.add(self.env.docname)  # type: ignore[attr-defined]
        return [node]


def html_moldoc(self: HTML5Translator, node: MolDocNode) -> None:
    molecule = node.get_molecule()

    default_molecule_config = self.config.moldoc_default_molecule_config
    if (config := molecule.get_config()) is not None:
        molecule_config = default_molecule_config.update(config)
    else:
        molecule_config = default_molecule_config

    moldoc_node_id = node.get_moldoc_name()

    if not getattr(self, "moldoc_scripts_added", False):
        self.body.append("<script>let moldoc_molecules=[];</script>")
        self.moldoc_scripts_added = True

    content = (
        "moldoc_molecules.push(() => {"
        "const md = molDraw;"
        f"let atoms={get_atom_array(molecule.get_atoms())};"
        f"let bonds={get_bond_array(molecule.get_bonds())};"
        "let maybeMolecule=md.maybeMolecule(atoms)(bonds);"
        "if (md.isLeft(maybeMolecule))"
        "{"
        'console.log("There was an issue with your molecule.");'
        "console.log(md.fromLeft()(maybeMolecule));"
        "}"
        "else"
        "{"
        "const molecule=md.fromRight()(maybeMolecule);"
        "const scene=md.scene({"
        f"{get_scene_config(moldoc_node_id, molecule_config)}"
        "});"
        "const meshes=md.meshes({"
        f"{get_mesh_config(molecule, molecule_config)}"
        "})(molecule);"
        "md.drawMol(scene(meshes));"
        "}"
        "});"
    )

    self.body.append(
        f'<div id="{moldoc_node_id}" style="height:25vh;"></div>'
        f"<script>{content}</script>"
    )
    raise nodes.SkipNode


def copy_asset_files(app: Sphinx, exc: Exception | None) -> None:
    asset_files = (
        "molDraw.js",
        "three.min.js",
    )
    static_dir = app.builder.outdir / "_static"
    # Build is HTML and succeeded.
    if app.builder.format == "html" and exc is None:
        for path in asset_files:
            with static_dir.joinpath(path).open("wb") as f:
                if (data := pkgutil.get_data(__package__, path)) is not None:
                    f.write(data)
                else:
                    msg = f"{path} not found"
                    raise RuntimeError(msg)


def add_moldoc_scripts(
    app: Sphinx,
    pagename: str,
    templatename: str,  # noqa: ARG001
    context: dict[str, typing.Any],  # noqa: ARG001
    doctree: nodes.document,  # noqa: ARG001
) -> None:
    if (
        hasattr(app.env, "moldoc_documents")
        and pagename in app.env.moldoc_documents
    ):
        app.add_js_file("three.min.js")
        app.add_js_file("molDraw.js")
        app.add_js_file(None, body="moldoc_molecules.forEach(f=>f());")


def purge_moldoc_documents(
    app: Sphinx,  # noqa: ARG001
    env: BuildEnvironment,
    docnames: list[str],
) -> None:
    if hasattr(env, "moldoc_documents"):
        env.moldoc_documents.difference_update(docnames)


def merge_moldoc_documents(
    app: Sphinx,  # noqa: ARG001
    env: BuildEnvironment,
    docnames: list[str],  # noqa: ARG001
    other: BuildEnvironment,
) -> None:
    if not hasattr(env, "moldoc_documents"):
        env.moldoc_documents = set()  # type: ignore[attr-defined]
    if hasattr(other, "moldoc_documents"):
        env.moldoc_documents.update(other.moldoc_documents)  # type: ignore[attr-defined]


def setup(app: Sphinx) -> dict:
    app.connect("build-finished", copy_asset_files)
    app.connect("html-page-context", add_moldoc_scripts)
    app.connect("env-purge-doc", purge_moldoc_documents)
    app.connect("env-merge-info", merge_moldoc_documents)
    app.add_directive("moldoc", MolDoc)
    app.add_node(
        node=MolDocNode,
        html=(html_moldoc, None),
    )
    app.add_config_value(
        "moldoc_default_molecule_config",
        MoleculeConfig(),
        "html",
        [MoleculeConfig],
    )
    return {
        "version": __version__,
        "parallel_read_safe": True,
        "parallel_write_safe": True,
    }



from collections.abc import Iterable
from typing import TypeAlias


class Color:
    def __init__(self, red: int, green: int, blue: int) -> None:
        self._red = red
        self._green = green
        self._blue = blue

    def get_red(self) -> int:
        return self._red

    def get_green(self) -> int:
        return self._green

    def get_blue(self) -> int:
        return self._blue


class AtomConfig:
    def __init__(
        self,
        color: Color | None = None,
        size: float | None = None,
    ) -> None:
        self._color = color
        self._size = size

    def get_color(self) -> Color | None:
        return self._color

    def get_size(self) -> float | None:
        return self._size


class Atom:
    def __init__(
        self,
        atomic_number: int,
        position: tuple[float, float, float],
        config: AtomConfig | None = None,
    ) -> None:
        self._atomic_number = atomic_number
        self._position = position
        self._config = config

    def get_atomic_number(self) -> int:
        return self._atomic_number

    def get_position(self) -> tuple[float, float, float]:
        return self._position

    def get_config(self) -> AtomConfig | None:
        return self._config


class Bond:
    def __init__(
        self,
        atom1_id: int,
        atom2_id: int,
        order: int,
    ) -> None:
        self._atom1_id = atom1_id
        self._atom2_id = atom2_id
        self._order = order

    def get_atom1_id(self) -> int:
        return self._atom1_id

    def get_atom2_id(self) -> int:
        return self._atom2_id

    def get_order(self) -> int:
        return self._order


class MeshLambertMaterial:
    pass


class MeshNormalMaterial:
    pass


class MeshPhongMaterial:
    def __init__(
        self,
        shininess: float | None = None,
        reflectivity: float | None = None,
        flat_shading: bool | None = None,
    ) -> None:
        self._shininess = shininess
        self._reflectivity = reflectivity
        self._flat_shading = flat_shading

    def get_shininess(self) -> float | None:
        return self._shininess

    def get_reflectivity(self) -> float | None:
        return self._reflectivity

    def get_flat_shading(self) -> bool | None:
        return self._flat_shading


class MeshPhysicalMaterial:
    def __init__(  # noqa: PLR0913
        self,
        clearcoat: float | None = None,
        clearcoat_roughness: float | None = None,
        roughness: float | None = None,
        metalness: float | None = None,
        reflectivity: float | None = None,
        flat_shading: bool | None = None,
    ) -> None:
        self._clearcoat = clearcoat
        self._clearcoat_roughness = clearcoat_roughness
        self._roughness = roughness
        self._metalness = metalness
        self._reflectivity = reflectivity
        self._flat_shading = flat_shading

    def get_clearcoat(self) -> float | None:
        return self._clearcoat

    def get_clearcoat_roughness(self) -> float | None:
        return self._clearcoat_roughness

    def get_roughness(self) -> float | None:
        return self._roughness

    def get_metalness(self) -> float | None:
        return self._metalness

    def get_reflectivity(self) -> float | None:
        return self._reflectivity

    def get_flat_shading(self) -> bool | None:
        return self._flat_shading


class MeshStandardMaterial:
    def __init__(
        self,
        metalness: float | None = None,
        roughness: float | None = None,
        flat_shading: bool | None = None,
    ) -> None:
        self._metalness = metalness
        self._roughness = roughness
        self._flat_shading = flat_shading

    def get_metalness(self) -> float | None:
        return self._metalness

    def get_roughness(self) -> float | None:
        return self._roughness

    def get_flat_shading(self) -> bool | None:
        return self._flat_shading


Material: TypeAlias = (
    MeshLambertMaterial
    | MeshNormalMaterial
    | MeshPhongMaterial
    | MeshPhysicalMaterial
    | MeshStandardMaterial
)


class MoleculeConfig:
    def __init__(
        self,
        atom_scale: float | None = None,
        material: Material | None = None,
        background_color: Color | None = None,
        is_outlined: bool | None = None,
    ) -> None:
        self._atom_scale = atom_scale
        self._material = material
        self._background_color = background_color
        self._is_outlined = is_outlined

    def get_atom_scale(self) -> float | None:
        return self._atom_scale

    def get_material(self) -> Material | None:
        return self._material

    def get_background_color(self) -> Color | None:
        return self._background_color

    def is_outlined(self) -> bool | None:
        return self._is_outlined

    def update(self, other: "MoleculeConfig") -> "MoleculeConfig":
        atom_scale = (
            self._atom_scale
            if other.get_atom_scale() is None
            else other.get_atom_scale()
        )
        material = (
            self._material
            if other.get_material() is None
            else other.get_material()
        )
        background_color = (
            self._background_color
            if other.get_background_color() is None
            else other.get_background_color()
        )
        is_outlined = (
            self._is_outlined
            if other.is_outlined() is None
            else other.is_outlined()
        )
        return MoleculeConfig(
            atom_scale=atom_scale,
            material=material,
            background_color=background_color,
            is_outlined=is_outlined,
        )


class Molecule:
    def __init__(
        self,
        atoms: Iterable[Atom],
        bonds: Iterable[Bond],
        config: MoleculeConfig | None = None,
    ) -> None:
        self._atoms = tuple(atoms)
        self._bonds = tuple(bonds)
        self._config = config

    def get_atoms(self) -> tuple[Atom, ...]:
        return self._atoms

    def get_bonds(self) -> tuple[Bond, ...]:
        return self._bonds

    def get_config(self) -> MoleculeConfig | None:
        return self._config



from typing import TypeVar

from moldoc._internal.javascript.bool import bool_to_javascript
from moldoc.molecule import (
    Material,
    MeshLambertMaterial,
    MeshNormalMaterial,
    MeshPhongMaterial,
    MeshPhysicalMaterial,
    MeshStandardMaterial,
)


def material_to_javascript(material: Material) -> str:
    match material:
        case MeshLambertMaterial():
            return (
                "material:(color)=>"
                "new THREE.MeshLambertMaterial({color:color})"
            )
        case MeshNormalMaterial():
            return "material:(color)=>new THREE.MeshNormalMaterial()"
        case MeshPhongMaterial():
            return _phong_to_javascript(material)
        case MeshPhysicalMaterial():
            return _physical_to_javascript(material)
        case MeshStandardMaterial():
            return _standard_to_javascript(material)


def _phong_to_javascript(material: MeshPhongMaterial) -> str:
    shininess = _property_to_javascript(
        property_name="shininess",
        property_value=material.get_shininess(),
    )
    reflectivity = _property_to_javascript(
        property_name="reflectivity",
        property_value=material.get_reflectivity(),
    )
    flat_shading = _property_to_javascript(
        property_name="flatShading",
        property_value=material.get_flat_shading(),
    )
    return (
        "material:"
        "(color)=>new THREE.MeshPhongMaterial({"
        "color:color"
        f'{"," if shininess else ""}{shininess}'
        f'{"," if reflectivity else ""}{reflectivity}'
        f'{"," if flat_shading else ""}{flat_shading}'
        "})"
    )


def _physical_to_javascript(material: MeshPhysicalMaterial) -> str:
    clearcoat = _property_to_javascript(
        property_name="clearcoat",
        property_value=material.get_clearcoat(),
    )
    clearcoat_roughness = _property_to_javascript(
        property_name="clearcoat_roughness",
        property_value=material.get_clearcoat_roughness(),
    )
    roughness = _property_to_javascript(
        property_name="roughness",
        property_value=material.get_roughness(),
    )
    metalness = _property_to_javascript(
        property_name="metalness",
        property_value=material.get_metalness(),
    )
    reflectivity = _property_to_javascript(
        property_name="reflectivity",
        property_value=material.get_reflectivity(),
    )
    flat_shading = _property_to_javascript(
        property_name="flat_shading",
        property_value=material.get_flat_shading(),
    )
    return (
        "material:"
        "(color)=>new THREE.MeshPhysicalMaterial({"
        "color:color"
        f'{"," if clearcoat else ""}{clearcoat}'
        f'{"," if clearcoat_roughness else ""}{clearcoat_roughness}'
        f'{"," if roughness else ""}{roughness}'
        f'{"," if metalness else ""}{metalness}'
        f'{"," if reflectivity else ""}{reflectivity}'
        f'{"," if flat_shading else ""}{flat_shading}'
        "})"
    )


def _standard_to_javascript(material: MeshStandardMaterial) -> str:
    metalness = _property_to_javascript(
        property_name="metalness",
        property_value=material.get_metalness(),
    )
    roughness = _property_to_javascript(
        property_name="roughness",
        property_value=material.get_roughness(),
    )
    flat_shading = _property_to_javascript(
        property_name="flatShading",
        property_value=material.get_flat_shading(),
    )
    return (
        "material:"
        "(color)=>new THREE.MeshStandardMaterial({"
        "color:color"
        f'{"," if metalness else ""}{metalness}'
        f'{"," if roughness else ""}{roughness}'
        f'{"," if flat_shading else ""}{flat_shading}'
        "})"
    )


T = TypeVar("T")


def _property_to_javascript(
    property_name: str,
    property_value: T | None,
) -> str:
    if property_value is None:
        return ""

    if isinstance(property_value, bool):
        return f"{property_name}:{bool_to_javascript(property_value)}"

    return f"{property_name}:{property_value}"



from collections.abc import Iterable

from moldoc.molecule import Bond


def get_bond_array(
    bonds: Iterable[Bond],
) -> str:
    bonds_js = ",".join(map(_bond_to_javascript, bonds))
    return f"[{bonds_js}]"


def _bond_to_javascript(bond: Bond) -> str:
    return (
        "md.bond"
        f"({bond.get_order()})"
        f"({bond.get_atom1_id()})"
        f"({bond.get_atom2_id()})"
    )



from moldoc.molecule import Color


def color_to_javascript(color: Color) -> str:
    red = f'{format(color.get_red(), "x")}'.rjust(2, "0")
    green = f'{format(color.get_green(), "x")}'.rjust(2, "0")
    blue = f'{format(color.get_blue(), "x")}'.rjust(2, "0")
    return f"0x{red}{green}{blue}"



from collections.abc import Iterable

from moldoc._internal.javascript.color import color_to_javascript
from moldoc._internal.javascript.material import material_to_javascript
from moldoc.molecule import Atom, Molecule, MoleculeConfig


def get_mesh_config(
    molecule: Molecule,
    molecule_config: MoleculeConfig | None,
) -> str:
    mesh_config = []
    atoms = molecule.get_atoms()

    atom_size_function = _get_atom_size_function(atoms)
    if atom_size_function is not None:
        mesh_config.append(atom_size_function)

    atom_color_function = _get_atom_color_function(atoms)
    if atom_color_function is not None:
        mesh_config.append(atom_color_function)

    if molecule_config is not None:
        atom_scale = molecule_config.get_atom_scale()
        if atom_scale is not None:
            mesh_config.append(f"atomScale:{atom_scale}")

        material = molecule_config.get_material()
        if material is not None:
            mesh_config.append(material_to_javascript(material))

    return ",".join(mesh_config)


def _get_atom_size_function(
    atoms: Iterable[Atom],
) -> str | None:
    atom_size_cases = tuple(_get_atom_size_cases(atoms))
    if atom_size_cases:
        return (
            "atomSize:atom=>{switch(md.id(atom)){"
            f'{"".join(atom_size_cases)}'
            "default:return md.size(md.chemicalSymbol(atom));"
            "}}"
        )

    return None


def _get_atom_size_cases(
    atoms: Iterable[Atom],
) -> Iterable[str]:
    for atom_id, atom in enumerate(atoms):
        config = atom.get_config()
        if config is not None:
            size = config.get_size()
            if size is not None:
                yield f"case {atom_id}:return {size};"


def _get_atom_color_function(
    atoms: Iterable[Atom],
) -> str | None:
    atom_color_cases = tuple(_get_atom_color_cases(atoms))
    if atom_color_cases:
        return (
            "atomColor:atom=>{switch(md.id(atom)){"
            f'{"".join(atom_color_cases)}'
            "default:return md.color(md.chemicalSymbol(atom));"
            "}}"
        )
    return None


def _get_atom_color_cases(
    atoms: Iterable[Atom],
) -> Iterable[str]:
    for atom_id, atom in enumerate(atoms):
        config = atom.get_config()
        if config is not None:
            color = config.get_color()
            if color is not None:
                color_js = color_to_javascript(color)
                yield f"case {atom_id}:return {color_js};"


def _molecule_config_to_javascript(
    config: MoleculeConfig | None,
) -> str | None:
    if config is None:
        return None

    javascript = []

    material = config.get_material()
    if material is not None:
        javascript.append(material_to_javascript(material))

    if javascript:
        return ",".join(javascript)

    return None
